@@ -6,10 +6,10 @@ import numpy as np
 import matplotlib.pyplot as plt
 
 from typing import *
-from scipy import interp
 from scipy import stats as ss
 from sklearn import metrics
 from functools import partial
+from abc import ABC, abstractmethod
 
 from ..misc import *
 
@@ -118,12 +118,11 @@ class Metrics(LoggingMixin):
     """
 
     sign_dict = {
-        "ddr": -1, "quantile": -1, "cdf_loss": -1, "loss": -1,
-        "f1_score": 1, "r2_score": 1, "auc": 1, "multi_auc": 1,
+        "f1_score": 1, "r2_score": 1, "auc": 1,
         "acc": 1, "mae": -1, "mse": -1, "ber": -1,
-        "correlation": 1, "top_k_score": 1
+        "quantile": -1, "cdf_loss": -1, "correlation": 1
     }
-    requires_prob_metrics = {"auc", "multi_auc"}
+    requires_prob_metrics = {"auc"}
     optimized_binary_metrics = {"acc", "ber"}
     custom_metrics = {}
 
@@ -225,21 +224,7 @@ class Metrics(LoggingMixin):
         n_classes = pred.shape[1]
         if n_classes == 2:
             return metrics.roc_auc_score(y.ravel(), pred[..., 1])
-        return Metrics.multi_auc(y, pred)
-
-    @staticmethod
-    def multi_auc(y, pred):
-        n_classes = pred.shape[1]
-        y = get_one_hot(y.ravel(), n_classes)
-        fpr, tpr = [None] * n_classes, [None] * n_classes
-        for i in range(n_classes):
-            fpr[i], tpr[i], _ = metrics.roc_curve(y[..., i], pred[..., i])
-        new_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))
-        new_tpr = np.zeros_like(new_fpr)
-        for i in range(n_classes):
-            new_tpr += interp(new_fpr, fpr[i], tpr[i])
-        new_tpr /= n_classes
-        return metrics.auc(new_fpr, new_tpr)
+        return metrics.roc_auc_score(y.ravel(), pred, multi_class="ovr")
 
     @staticmethod
     def acc(y, pred):
@@ -284,8 +269,17 @@ class Metrics(LoggingMixin):
         return thresholds[np.argmax(metric)]
 
 
+def register_metric(name, sign, requires_prob):
+    def _register(f):
+        Metrics.add_metric(f, name, sign, requires_prob)
+        return f
+    return _register
+
+
 estimate_fn_type = Callable[[np.ndarray], np.ndarray]
 scoring_fn_type = Callable[[List[float], float, float], float]
+collate_fn_type = Callable[[List[np.ndarray], bool], np.ndarray]
+predict_method_type = Union[estimate_fn_type, None]
 
 
 class Estimator(LoggingMixin):
@@ -360,7 +354,7 @@ class Estimator(LoggingMixin):
                  methods: Dict[str, Union[estimate_fn_type, List[estimate_fn_type]]],
                  *,
                  scoring_function: Union[str, scoring_fn_type] = "default",
-                 verbose_level: int = 1) -> None:
+                 verbose_level: int = 1) -> Dict[str, Dict[str, float]]:
         self._reset()
         if isinstance(scoring_function, str):
             scoring_function = getattr(self, f"_{scoring_function}_scoring")
@@ -372,6 +366,7 @@ class Estimator(LoggingMixin):
             for name, sub_methods in methods.items()
         }
         msg_list = []
+        statistics = {}
         best_idx, best_score = -1, -math.inf
         sorted_method_names = sorted(self.raw_metrics)
         for i, name in enumerate(sorted_method_names):
@@ -382,15 +377,35 @@ class Estimator(LoggingMixin):
             self.final_scores[name] = new_score
             if new_score > best_score:
                 best_idx, best_score = i, new_score
+            method_statistics = statistics.setdefault(name, {})
+            method_statistics["mean"], method_statistics["std"] = mean, std
+            method_statistics["score"] = new_score
         self.best_method = sorted_method_names[best_idx]
         msg_list[best_idx] += "  <-  "
         width = max(map(len, msg_list))
         msg_list.insert(0, "=" * width)
         msg_list.append("-" * width)
         self.log_block_msg("\n".join(msg_list), self.info_prefix, "Results", verbose_level)
+        return statistics
 
 
-class ModelPattern(LoggingMixin):
+class PatternBase(ABC):
+    @abstractmethod
+    def predict_method(self,
+                       requires_prob: bool) -> predict_method_type:
+        pass
+
+    def predict(self,
+                x: np.ndarray,
+                *,
+                requires_prob: bool = False) -> np.ndarray:
+        predict_method = self.predict_method(requires_prob)
+        if predict_method is None:
+            raise ValueError(f"predicting with requires_prob={requires_prob} is not defined")
+        return predict_method(x)
+
+
+class ModelPattern(PatternBase, LoggingMixin):
     """
     Util class to create an interface for users to leverage `Comparer` & `HPO` (and more in the future).
 
@@ -447,12 +462,12 @@ class ModelPattern(LoggingMixin):
         self._verbose_level = verbose_level
 
     def predict_method(self,
-                       requires_prob: bool) -> Callable[[np.ndarray], np.ndarray]:
+                       requires_prob: bool) -> predict_method_type:
         predict_method = self._predict_prob_method if requires_prob else self._predict_method
         if isinstance(predict_method, str):
             if self.model is None:
                 raise ValueError("Either init_method or Callable predict_method is required in ModelPattern")
-            predict_method = getattr(self.model, predict_method)
+            predict_method = getattr(self.model, predict_method, None)
         elif self.model is not None:
             self.log_msg(
                 "predict_method is Callable but model is also created, which has no effect",
@@ -460,21 +475,12 @@ class ModelPattern(LoggingMixin):
             )
         return predict_method
 
-    def predict(self,
-                x: np.ndarray,
-                *,
-                requires_prob: bool = False) -> np.ndarray:
-        return self.predict_method(requires_prob)(x)
-
     @classmethod
     def repeat(cls, n: int, **kwargs) -> List["ModelPattern"]:
         return [cls(**kwargs) for _ in range(n)]
 
 
-collate_fn_type = Callable[[List[np.ndarray], bool], np.ndarray]
-
-
-class EnsemblePattern:
+class EnsemblePattern(PatternBase):
     """
     Util class to create an interface for users to leverage `Comparer` & `HPO` in an ensembled way.
 
@@ -536,19 +542,16 @@ class EnsemblePattern:
     # API
 
     def predict_method(self,
-                       requires_prob: bool) -> Callable[[np.ndarray], np.ndarray]:
+                       requires_prob: bool) -> predict_method_type:
         predict_methods = list(map(ModelPattern.predict_method, self._patterns, len(self) * [requires_prob]))
+        predict_methods = [method for method in predict_methods if method is not None]
+        if not predict_methods:
+            return
         def _predict(x: np.ndarray):
             predictions = [method(x) for method in predict_methods]
             return self.collate_fn(predictions, requires_prob)
         return _predict
 
-    def predict(self,
-                x: np.ndarray,
-                *,
-                requires_prob: bool = False) -> np.ndarray:
-        return self.predict_method(requires_prob)(x)
-
     @classmethod
     def from_same_methods(cls,
                           n: int,
@@ -559,6 +562,7 @@ class EnsemblePattern:
 
 pattern_type = Union[ModelPattern, EnsemblePattern]
 patterns_type = Union[pattern_type, List[pattern_type]]
+choices_type = Union[None, List[Union[int, Set[int], None]]]
 
 
 class Comparer(LoggingMixin):
@@ -602,7 +606,7 @@ class Comparer(LoggingMixin):
     >>> comparer = Comparer(patterns, estimators).compare(x, y)
     >>> # {'mse': {'identical': 1.666667, 'minus_one': 0.666666},
     >>> # 'mae': {'identical': 1.0, 'minus_one': 0.666666}}
-    >>> print(comparer.scores)
+    >>> print(comparer.final_scores)
     >>> # {'mse': 'minus_one', 'mae': 'minus_one'}
     >>> print(comparer.best_methods)
 
@@ -629,12 +633,114 @@ class Comparer(LoggingMixin):
     def best_methods(self) -> Dict[str, str]:
         return {k: v.best_method for k, v in self.estimators.items()}
 
+    def _log_statistics(self,
+                        verbose_level: int,
+                        **kwargs) -> None:
+        sorted_metrics = sorted(self.estimator_statistics)
+        body = {}
+        same_choices: choices_type = None
+        best_choices: choices_type = None
+        need_display_best_choice = False
+        sub_header = sorted_methods = None
+        statistic_types = ["mean", "std", "score"]
+        for metric_idx, metric_type in enumerate(sorted_metrics):
+            statistics = self.estimator_statistics[metric_type]
+            if sorted_methods is None:
+                sorted_methods = sorted(statistics)
+                need_display_best_choice = len(sorted_methods) > 1
+            if sub_header is None:
+                sub_header = statistic_types * len(sorted_metrics)
+            if best_choices is None and need_display_best_choice:
+                same_choices = [None] * len(sub_header)
+                best_choices = [None] * len(sub_header)
+            for method_idx, method in enumerate(sorted_methods):
+                method_statistics = statistics[method]
+                method_statistics = [method_statistics[stat_type] for stat_type in statistic_types]
+                body.setdefault(method, []).extend(method_statistics)
+                if best_choices is not None:
+                    for statistic_idx, method_statistic in enumerate(method_statistics):
+                        choice_idx = metric_idx * len(statistic_types) + statistic_idx
+                        current_idx_choice = best_choices[choice_idx]
+                        if current_idx_choice is None:
+                            best_choices[choice_idx] = method_idx
+                        else:
+                            stat_type = statistic_types[statistic_idx]
+                            chosen_stat = statistics[sorted_methods[current_idx_choice]][stat_type]
+                            if method_statistic == chosen_stat:
+                                if same_choices[choice_idx] is None:
+                                    same_choices[choice_idx] = {method_idx}
+                                else:
+                                    same_choices[choice_idx].add(method_idx)
+                            elif stat_type == "std":
+                                if method_statistic < chosen_stat:
+                                    same_choices[choice_idx] = None
+                                    best_choices[choice_idx] = method_idx
+                            elif stat_type == "score":
+                                if method_statistic > chosen_stat:
+                                    same_choices[choice_idx] = None
+                                    best_choices[choice_idx] = method_idx
+                            else:
+                                assert stat_type == "mean"
+                                sign = Metrics.sign_dict[metric_type]
+                                if method_statistic * sign > chosen_stat * sign:
+                                    same_choices[choice_idx] = None
+                                    best_choices[choice_idx] = method_idx
+        padding = 2 * (kwargs.get("padding", 1) + 3)
+        method_length = kwargs.get("method_length", 16) + padding
+        float_length = kwargs.get("float_length", 8)
+        cell_length = float_length + padding
+        num_statistic_types = len(statistic_types)
+        metric_type_length = num_statistic_types * cell_length + 2
+        header_msg = (
+            f"|{'metrics':^{method_length}s}|"
+            + "|".join([
+                f"{metric_type:^{metric_type_length}s}"
+                for metric_type in sorted_metrics
+            ])
+            + "|"
+        )
+        sub_header_msg = (
+            f"|{' ' * method_length}|"
+            + "|".join([f"{sub_header_item:^{cell_length}s}" for sub_header_item in sub_header])
+            + "|"
+        )
+        body_msgs = []
+        for method_idx, method in enumerate(sorted_methods):
+            cell_msgs = []
+            for cell_idx, cell_item in enumerate(body[method]):
+                cell_str = fix_float_to_length(cell_item, float_length)
+                if (
+                    best_choices is not None
+                    and (best_choices[cell_idx] == method_idx
+                         or same_choices[cell_idx] is not None and method_idx in same_choices[cell_idx])
+                ):
+                    cell_str = f" -- {cell_str} -- "
+                else:
+                    cell_str = f"{cell_str:^{cell_length}s}"
+                cell_msgs.append(cell_str)
+            body_msgs.append(
+                f"|{method:^{method_length}s}|"
+                + "|".join(cell_msgs)
+                + "|"
+            )
+        msgs = [header_msg, sub_header_msg] + body_msgs
+        length = len(body_msgs[0])
+        single_split = "-" * length
+        double_split = "=" * length
+        main_msg = f"\n{single_split}\n".join(msgs)
+        self.log_block_msg(
+            f"{double_split}\n{main_msg}\n{double_split}",
+            self.info_prefix, "Results", verbose_level
+        )
+
     def compare(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 *,
                 scoring_function: Union[str, scoring_fn_type] = "default",
-                verbose_level: int = 1) -> "Comparer":
+                verbose_level: int = 1,
+                **kwargs) -> "Comparer":
+        self.estimator_statistics: Dict[str, Dict[str, Dict[str, float]]] = {}
         for estimator in self.estimators.values():
             methods = {}
             for model_name, patterns in self.patterns.items():
@@ -663,11 +769,12 @@ class Comparer(LoggingMixin):
                 if invalid:
                     continue
                 methods[model_name] = predict_methods
-            estimator.estimate(
+            self.estimator_statistics[estimator.type] = estimator.estimate(
                 x, y, methods,
                 scoring_function=scoring_function,
-                verbose_level=verbose_level
+                verbose_level=verbose_level + 5
             )
+        self._log_statistics(verbose_level, **kwargs)
         return self
 
 
@@ -909,5 +1016,7 @@ class Visualizer:
 __all__ = [
     "Anneal", "Metrics", "ScalarEMA", "Visualizer",
     "Estimator", "ModelPattern", "EnsemblePattern", "Comparer",
-    "pattern_type", "patterns_type", "estimate_fn_type", "scoring_fn_type"
+    "collate_fn_type", "estimate_fn_type", "scoring_fn_type",
+    "pattern_type", "patterns_type",
+    "register_metric"
 ]
